package com.ijchart.xychart.marker ;

import java.awt.BasicStroke ;
import java.awt.Color ;
import java.awt.Font ;
import java.awt.Paint ;
import java.awt.Stroke ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;
import java.util.EventListener ;

import javax.swing.event.EventListenerList ;

import com.ijchart.xychart.constant.HorizontalAlignment ;
import com.ijchart.xychart.constant.TextPosition ;
import com.ijchart.xychart.constant.VerticalAlignment ;
import com.ijchart.xychart.event.MarkerChangeEvent ;
import com.ijchart.xychart.listener.IMarkerChangeListener ;
import com.ijchart.xychart.ui.RectangleInsets ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;
import com.ijchart.xychart.util.SerialUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ͼеĳЩcategoryvalueֵҪصʾ,ǻ.</zh>
 * <en>The base class for markers that can be added to plots to highlight a value
 * or range of values.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public abstract class AbstractMarker implements Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -7478942750107485276L ;

  /**
   * <zh>ȱʡʹõPaint</zh>
   * <en>The default fill Paint</en>
   */
  public static final transient Paint DEFAULT_FILL_PAINT = new Color(221, 255, 221, 128) ;

  /**
   * <zh>ȱʡֺͼελùϵ.</zh>
   * <en>The default text position.</en>
   */
  public static final TextPosition DEFAULT_TEXT_POSITION = TextPosition.INSIDE ;

  /**
   * <zh>Ƿͼڲ</zh>
   * <en>Whether or not fill ChartShape</en>
   */
  private boolean fillShape ;

  /**
   * <zh>ʹõPaint</zh>
   * <en>The fill Paint</en>
   */
  private transient Paint fillPaint ;

  /**
   * <zh>䱳alphaֵ.</zh>
   * <en>The alpha value for fill background.</en>
   */
  private float fillAlpha ;

  /**
   * <zh>Ƿ.</zh>
   * <en>Whether or not draw border.</en>
   */
  private boolean drawBorder ;

  /**
   * <zh>ʹõPaint</zh>
   * <en>The border Paint</en>
   */
  private transient Paint borderPaint ;

  /**
   * <zh>ʹõStroke</zh>
   * <en>The border Stroke</en>
   */
  private transient Stroke borderStroke ;

  /**
   * <zh>ʾ.</zh>
   * <en>The text.</en>
   */
  private String text ;

  /**
   * <zh>߿֮Ŀհ.</zh>
   * <en>The margin around the inside of the draw area.</en>
   */
  private RectangleInsets margin ;

  /**
   * <zh>Ƿʾ.</zh>
   * <en>A flag that controls whether or not the text is drawn.</en>
   */
  private boolean drawText ;

  /**
   * <zh>ֵ.</zh>
   * <en>The text font.</en>
   */
  private Font textFont ;

  /**
   * <zh>ڲʹõPaint</zh>
   * <en>The data Paint</en>
   */
  private transient Paint textPaint ;

  /**
   * <zh>ˮƽ뷽ʽ.</zh>
   * <en>The text horizontal align.</en>
   */
  private HorizontalAlignment textHorizontalAlign ;

  /**
   * <zh>ֱִ뷽ʽ.</zh>
   * <en>The text vertical align.</en>
   */
  private VerticalAlignment textVerticalAlign ;

  /**
   * <zh>ıλ</zh>
   * <en>The text position.</en>
   */
  private TextPosition textPosition ;

  /**
   * <zh>,ɱл.</zh>
   * <en>The listener collection,not serializable.</en>
   */
  private transient EventListenerList listenerList ;

  /**
   * <zh>ȱʡĹ.</zh>
   * <en>Creates a new marker with default attributes.</en>
   */
  protected AbstractMarker() {
    this(DEFAULT_FILL_PAINT) ;
  }

  /**
   * <zh>һ¶.</zh>
   * <en>Constructs a new marker.</en>
   *
   * @param fillPaint <zh>ʹõPaint (Ϊnull).</zh>
   * <en>The fill paint (null not permitted).</en>
   */
  protected AbstractMarker(Paint fillPaint) {
    this(fillPaint, Color.gray, new BasicStroke(1f), 0.80f) ;
  }

  /**
   * <zh>ʹв.</zh>
   * <en>Constructs a new marker.</en>
   *
   * @param fillPaint Paint <zh>ʹõpaint (Ϊnull).</zh>
   * <en>The fill paint (null not permitted).</en>
   * @param borderPaint Paint <zh>ʹõPaint (Ϊnull).</zh>
   * <en>The border paint (null permitted).</en>
   * @param borderStroke Stroke <zh>ʹõStroke (Ϊnull).</zh>
   * <en>The border stroke (null permitted).</en>
   * @param fillAlpha float <zh>䱳alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (must be in the range 0.0f to 1.0f).</en>
   */
  protected AbstractMarker(Paint fillPaint, Paint borderPaint, Stroke borderStroke,
                           float fillAlpha) {
    if (fillPaint == null) {
      throw new IllegalArgumentException("Null 'fillPaint' argument.") ;
    }
    if (borderPaint == null) {
      throw new IllegalArgumentException("Null 'borderPaint' argument.") ;
    }
    if (borderStroke == null) {
      throw new IllegalArgumentException("Null 'borderStroke' argument.") ;
    }
    if (fillAlpha < 0.0f || fillAlpha > 1.0f) {
      throw new IllegalArgumentException(
        "The 'fillAlpha' value must be in the range 0.0f to 1.0f") ;
    }

    this.fillShape = true ;
    this.fillPaint = fillPaint ;

    this.drawBorder = false ;
    this.borderPaint = borderPaint ;
    this.borderStroke = borderStroke ;
    this.fillAlpha = fillAlpha ;

    this.text = null ;
    this.textFont = new Font("SansSerif", Font.PLAIN, 9) ;
    this.textPaint = Color.black ;
    this.drawText = true ;

    this.margin = new RectangleInsets(3.0, 3.0, 3.0, 3.0) ;
    this.textHorizontalAlign = HorizontalAlignment.CENTER ;
    this.textVerticalAlign = VerticalAlignment.CENTER ;
    this.textPosition = DEFAULT_TEXT_POSITION ;

    this.listenerList = new EventListenerList() ;
  }

  /**
   * <zh>ǷƵı־</zh>
   * <en>A flag that controls whether or not the shape is filled</en>
   *
   * @return boolean <zh>Ƿtrue,򷵻false</zh>
   * <en>A boolean flag</en>
   */
  public boolean isFillShape() {
    return this.fillShape ;
  }

  /**
   * <zh>Ƿͼڲ,м{@link MarkerChangeEvent}¼.</zh>
   * <en>Sets a flag that controls whether or not the shape is filled,and sends a
   * {@link MarkerChangeEvent} to all registered listeners.</en>
   *
   * @param bFillShape boolean <zh>trueͼڲ,falseͼڲ</zh>
   * <en>true,fill the shape;false otherwise</en>
   */
  public void setFillShape(boolean bFillShape) {
    if (this.fillShape != bFillShape) {
      this.fillShape = bFillShape ;
      fireMarkerChanged() ;
    }
  }

  /**
   * <zh>ͼڲɫ</zh>
   * <en>Get the ChartShape fill paint</en>
   *
   * @return Paint <zh>ͼڲɫ</zh>
   * <en>The ChartShape fill paint</en>
   */
  public Paint getFillPaint() {
    return fillPaint ;
  }

  /**
   * <zh>ͼڲɫ,м{@link MarkerChangeEvent}¼.</zh>
   * <en>Sets the shape fill Paint,and sends a
   * {@link MarkerChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>ͼڲõPaint(Ϊnull)</zh>
   * <en>The ChartShape fill Paint(null not permitted).</en>
   */
  public void setFillPaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }
    this.fillPaint = paint ;
    fireMarkerChanged() ;
  }

  /**
   * <zh>䱳alphaֵ,м{@link MarkerChangeEvent}¼.</zh>
   * <en>The alpha value for fill background,and sends a
   * {@link MarkerChangeEvent} to all registered listeners.</en>
   *
   * @param alpha float <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public void setFillAlpha(float alpha) {
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException(
        "The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    if (this.fillAlpha != alpha) {
      this.fillAlpha = alpha ;
      fireMarkerChanged() ;
    }
  }

  /**
   * <zh>䱳alphaֵ.</zh>
   * <en>Returns the alpha value for fill background.</en>
   * @return float <zh>alphaֵ.</zh>
   * <en>The alpha value.</en>
   */
  public float getFillAlpha() {
    return this.fillAlpha ;
  }

  /**
   * <zh>Paint</zh>
   * <en>Get the ChartShape border Paint</en>
   *
   * @return Paint <zh>Paint</zh>
   * <en>The border Paint</en>
   */
  public Paint getBorderPaint() {
    return borderPaint ;
  }

  /**
   * <zh>ͼߵPaint,м{@link MarkerChangeEvent}¼.</zh>
   * <en>Sets the shape border Paint,and sends a
   * {@link MarkerChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>ͼPaint(Ϊnull)</zh>
   * <en>The ChartShape border Paint(null not permitted).</en>
   */
  public void setBorderPaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.borderPaint = paint ;
    fireMarkerChanged() ;
  }

  /**
   * <zh>߻</zh>
   * <en>Get the ChartShape border Stroke</en>
   *
   * @return Stroke <zh>߻</zh>
   * <en>The border Stroke</en>
   */
  public Stroke getBorderStroke() {
    return this.borderStroke ;
  }

  /**
   * <zh>ͼߵĻ,м{@link MarkerChangeEvent}¼.</zh>
   * <en>Sets the shape border Stroke,and sends a
   * {@link MarkerChangeEvent} to all registered listeners.</en>
   *
   * @param stroke Stroke <zh>ͼ߻(Ϊnull)</zh>
   * <en>The ChartShape border Stroke(null
   *   not permitted).</en>
   */
  public void setBorderStroke(Stroke stroke) {
    if (stroke == null) {
      throw new IllegalArgumentException("Null 'stroke' argument.") ;
    }

    this.borderStroke = stroke ;
    fireMarkerChanged() ;
  }

  /**
   * <zh>ǷƵı־</zh>
   * <en>A flag that controls whether or not the shape border is visible</en>
   *
   * @return boolean <zh>Ƿtrue,򷵻false</zh>
   * <en>A boolean flag</en>
   */
  public boolean isDrawBorder() {
    return drawBorder ;
  }

  /**
   * <zh>Ƿͼε,м{@link MarkerChangeEvent}¼.</zh>
   * <en>Sets a flag that controls whether or not the shape border is visible,and sends a
   * {@link MarkerChangeEvent} to all registered listeners.</en>
   *
   * @param flag boolean <zh>true,false</zh>
   * <en>true,draw the border of the shape;false otherwise</en>
   */
  public void setDrawBorder(boolean flag) {
    if (this.drawBorder != flag) {
      this.drawBorder = flag ;
      fireMarkerChanged() ;
    }
  }

  /**
   * <zh>ʾ,м{@link MarkerChangeEvent}¼.</zh>
   * <en>Sets the text,and sends a
   * {@link MarkerChangeEvent} to all registered listeners..</en>
   * @param text String <zh>Ҫʾ (Ϊnull).</zh>
   * <en>The text (null permitted).</en>
   */
  public void setText(String text) {
    if (!EqualsUtil.equals(this.text, text)) {
      this.text = text ;
      fireMarkerChanged() ;
    }
  }

  /**
   * <zh>ʾ.</zh>
   * <en>Returns the text.</en>
   *
   * @return String <zh>ʾ,Ϊnull.</zh>
   * <en>The text,maybe null.</en>
   */
  public String getText() {
    return this.text ;
  }

  /**
   * <zh>ַı߿֮Ŀհ,м{@link MarkerChangeEvent}¼.</zh>
   * <en>Sets the margin around the inside of the block,and sends a
   * {@link MarkerChangeEvent} to all registered listeners..</en>
   *
   * @param margin RectangleInsets <zh>ַı߿֮Ŀհ (Ϊnull).</zh>
   * <en>The margin around the inside of the block (null permitted).</en>
   */
  public void setMargin(RectangleInsets margin) {
    if (margin == null) {
      throw new IllegalArgumentException("Null 'margin' argument.") ;
    }
    this.margin = margin ;
    fireMarkerChanged() ;
  }

  /**
   * <zh>ַı߿֮Ŀհ.</zh>
   * <en>Returns the margin around the inside of the block.</en>
   *
   * @return RectangleInsets <zh>ַı߿֮Ŀհ.</zh>
   * <en>The margin around the inside of the block.</en>
   */
  public RectangleInsets getMargin() {
    return this.margin ;
  }

  /**
   * <zh>ֵ.</zh>
   * <en>Returns the text font.</en>
   *
   * @return Font
   */
  public Font getTextFont() {
    return this.textFont ;
  }

  /**
   * <zh>ֵ,
   * м{@link com.ijchart.xychart.event.MarkerChangeEvent}¼.</zh>
   * <en>Sets the text font and sends a {@link MarkerChangeEvent} to all registered listeners.</en>
   *
   * @param font Font <zh>ֵ (Ϊnull).</zh>
   * <en>The text font (null not permitted).</en>
   */
  public void setTextFont(Font font) {
    if (font == null) {
      throw new IllegalArgumentException("Null 'font' argument.") ;
    }

    if (!this.textFont.equals(font)) {
      this.textFont = font ;
      fireMarkerChanged() ;
    }
  }

  /**
   * <zh>Ƿʾ,
   * м{@link MarkerChangeEvent}¼.</zh>
   * <en>Sets a flag that controls whether or not the text is drawn and sends
   * an {@link MarkerChangeEvent} to all registered listeners.</en>
   *
   * @param flag boolean
   */
  public void setDrawText(boolean flag) {
    if (this.drawText != flag) {
      this.drawText = flag ;
      fireMarkerChanged() ;
    }
  }

  /**
   * <zh>Ƿʾ.</zh>
   * <en>Returns a flag that controls whether or not the text is drawn.</en>
   *
   * @return boolean
   */
  public boolean isDrawText() {
    return this.drawText ;
  }

  /**
   * <zh>ͼֵPaint,
   * м{@link MarkerChangeEvent}¼</zh>
   * <en>Sets the shape text Paint and sends
   * an {@link MarkerChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>ͼPaint(Ϊnull)</zh>
   *  <en>The ChartShape text Paint(null not permitted).</en>
   */
  public void setTextPaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.textPaint = paint ;
    fireMarkerChanged() ;
  }

  /**
   * <zh>Paint</zh>
   * <en>Get the ChartShape text Paint</en>
   *
   * @return Paint <zh>Paint</zh>
   * <en>The text Paint</en>
   */
  public Paint getTextPaint() {
    return this.textPaint ;
  }

  /**
   * <zh>ˮƽ뷽ʽ,
   * м{@link MarkerChangeEvent}¼.</zh>
   * <en>Sets the text horizontal alignment and sends
   *  an {@link MarkerChangeEvent} to all registered listeners.</en>
   *
   * @param horizontalAlign HorizontalAlignment <zh>ˮƽ뷽ʽ (Ϊnull).</zh>
   * <en>The text horizontal alignment (null not permitted).</en>
   */
  public void setTextHorizontalAlign(HorizontalAlignment horizontalAlign) {
    if (horizontalAlign == null) {
      throw new IllegalArgumentException("Null 'horizontalAlign' argument.") ;
    }

    if (!this.textHorizontalAlign.equals(horizontalAlign)) {
      this.textHorizontalAlign = horizontalAlign ;
      fireMarkerChanged() ;
    }
  }

  /**
   * <zh>ֵˮƽ뷽ʽ.</zh>
   * <en>Returns the text horizontal alignment.</en>
   *
   * @return HorizontalAlignment <zh>ˮƽ뷽ʽ.</zh>
   * <en>The text horizontal alignment.</en>
   */
  public HorizontalAlignment getTextHorizontalAlign() {
    return this.textHorizontalAlign ;
  }

  /**
   * <zh>ֱִ뷽ʽ,
   * м{@link MarkerChangeEvent}¼.</zh>
   * <en>Sets the text vertical alignment and sends
   * an {@link MarkerChangeEvent} to all registered listeners.</en>
   *
   * @param verticalAlign VerticalAlignment <zh>ֱִ뷽ʽ (Ϊnull).</zh>
   * <en>The text vertical alignment (null not permitted).</en>
   */
  public void setTextVerticalAlign(VerticalAlignment verticalAlign) {
    if (verticalAlign == null) {
      throw new IllegalArgumentException("Null 'verticalAlign' argument.") ;
    }

    if (!this.textVerticalAlign.equals(verticalAlign)) {
      this.textVerticalAlign = verticalAlign ;
      fireMarkerChanged() ;
    }
  }

  /**
   * <zh>ֵĴֱ뷽ʽ.</zh>
   * <en>Returns the text vertical alignment.</en>
   *
   * @return VerticalAlignment <zh>ֱִ뷽ʽ.</zh>
   * <en>The text vertical alignment.</en>
   */
  public VerticalAlignment getTextVerticalAlign() {
    return this.textVerticalAlign ;
  }

  /**
   * <zh>ֺͼελùϵ.</zh>
   * <en>Returns the text position.</en>
   *
   * @return TextPosition <zh>ֺͼελùϵ.</zh>
   * <en>The text position.</en>
   */
  public TextPosition getTextPosition() {
    return this.textPosition ;
  }

  /**
   * <zh>ֺͼελùϵ,
   * м{@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the text position and sends
   * a {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param position TextPosition <zh>ֺͼελùϵ (Ϊnull).</zh>
   * <en>The text position (null not permitted).</en>
   */
  public void setTextPosition(TextPosition position) {
    if (position == null) {
      throw new IllegalArgumentException("Null 'position' argument.") ;
    }

    if (!this.textPosition.equals(position)) {
      this.textPosition = position ;
      notifyListeners(new MarkerChangeEvent(this)) ;
    }
  }

  /**
   * <zh>һ</zh>
   * <en>Registers an object for notification of changes to the marker.</en>
   *
   * @param listener <zh>.</zh>
   * <en>The object to be registered.</en>
   */
  public void addChangeListener(IMarkerChangeListener listener) {
    this.listenerList.add(IMarkerChangeListener.class, listener) ;
  }

  /**
   * <zh>ɾһ.</zh>
   * <en>Unregisters an object for notification of changes to the marker.</en>
   *
   * @param listener <zh>Ҫɾļ.</zh>
   * <en>The object to be unregistered.</en>
   */
  public void removeChangeListener(IMarkerChangeListener listener) {
    this.listenerList.remove(IMarkerChangeListener.class, listener) ;
  }

  /**
   * <zh>֪ͨעļmarker˱仯.</zh>
   * <en>Notifies all registered listeners that the marker has been modified.</en>
   *
   * @param event <zh>¼.</zh>
   * <en>Information about the change event.</en>
   */
  protected void notifyListeners(MarkerChangeEvent event) {
    Object[] listeners = this.listenerList.getListenerList() ;
    for (int i = listeners.length - 2 ; i >= 0 ; i -= 2) {
      if (listeners[i] == IMarkerChangeListener.class) {
        ((IMarkerChangeListener) listeners[i + 1]).markerChanged(event) ;
      }
    }

  }

  /**
   * <zh>classеļ.</zh>
   * <en>Returns an array containing all the listeners of the specified type.</en>
   *
   * @param listenerType <zh>class.</zh>
   * <en>The listener type.</en>
   *
   * @return <zh>listenerType͵ļ.</zh>
   * <en>The array of listeners.</en>
   */
  public EventListener[] getListeners(Class listenerType) {
    return this.listenerList.getListeners(listenerType) ;
  }

  /**
   * <zh>ͨмݼ˱仯.</zh>
   * <en>Notifies all registered listeners that the dataset has changed.</en>
   */
  protected void fireMarkerChanged() {
    notifyListeners(new MarkerChangeEvent(this)) ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.  </en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "AbstractMarker" ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests the marker for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof AbstractMarker)) {
      return false ;
    }

    AbstractMarker that = (AbstractMarker) obj ;
    if (this.fillShape != that.fillShape) {
      return false ;
    }
    if (!EqualsUtil.equals(this.fillPaint, that.fillPaint)) {
      return false ;
    }
    if (this.drawBorder != that.drawBorder) {
      return false ;
    }
    if (!EqualsUtil.equals(this.borderPaint, that.borderPaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.borderStroke, that.borderStroke)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.text, that.text)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.textFont, that.textFont)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.textPaint, that.textPaint)) {
      return false ;
    }
    if (this.drawText != that.drawText) {
      return false ;
    }
    if (this.fillAlpha != that.fillAlpha) {
      return false ;
    }
    if (!EqualsUtil.equals(this.margin, that.margin)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.textHorizontalAlign, that.textHorizontalAlign)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.textVerticalAlign, that.textVerticalAlign)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.textPosition, that.textPosition)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int The hashcode value.
   */
  public int hashCode() {
    int result = HashcodeUtil.hashCode(this.fillShape) ;
    result = 37 * result + HashcodeUtil.hashCode(this.fillPaint) ;
    result = 37 * result + HashcodeUtil.hashCode(this.drawBorder) ;
    result = 37 * result + HashcodeUtil.hashCode(this.borderPaint) ;
    result = 37 * result + HashcodeUtil.hashCode(this.borderStroke) ;
    result = 37 * result + HashcodeUtil.hashCode(this.text) ;
    result = 37 * result + HashcodeUtil.hashCode(this.textFont) ;
    result = 37 * result + HashcodeUtil.hashCode(this.textPaint) ;
    result = 37 * result + HashcodeUtil.hashCode(this.drawText) ;
    result = 37 * result + (int)this.fillAlpha * 10 ;
    result = 37 * result + HashcodeUtil.hashCode(this.margin) ;
    result = 37 * result + HashcodeUtil.hashCode(this.textHorizontalAlign) ;
    result = 37 * result + HashcodeUtil.hashCode(this.textVerticalAlign) ;
    result = 37 * result + HashcodeUtil.hashCode(this.textPosition) ;

    return result ;
  }

  /**
   * <zh>Ʊ.</zh>
   * <en>Creates a clone of the marker.</en>
   *
   * @return <zh>ƺĶ</zh>
   * <en>A clone.</en>
   *
   * @throws CloneNotSupportedException never.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the output stream.
   *
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;
    SerialUtil.writePaint(stream, this.fillPaint) ;
    SerialUtil.writePaint(stream, this.borderPaint) ;
    SerialUtil.writeStroke(stream, this.borderStroke) ;
    SerialUtil.writePaint(stream, this.textPaint) ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;
    this.fillPaint = SerialUtil.readPaint(stream) ;
    this.borderPaint = SerialUtil.readPaint(stream) ;
    this.borderStroke = SerialUtil.readStroke(stream) ;
    this.textPaint = SerialUtil.readPaint(stream) ;

    this.listenerList = new EventListenerList() ;
  }

}
