package com.ijchart.xychart.render.category ;

import java.awt.geom.Rectangle2D ;
import java.io.Serializable ;
import java.util.ArrayList ;
import java.util.Collection ;
import java.util.Collections ;
import java.util.Iterator ;
import java.util.List ;

import com.ijchart.xychart.annotation.ICategoryAnnotation ;
import com.ijchart.xychart.axis.category.CategoryAxis ;
import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.chartinfo.AnnotationShapeCollection ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.marker.CategoryMarker ;
import com.ijchart.xychart.render.AbstractWithAxisRender ;
import com.ijchart.xychart.shape.IChartShape ;
import com.ijchart.xychart.shape.RectangleTextChartShape ;
import com.ijchart.xychart.ui.Layer ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;
import com.ijchart.xychart.util.StrokeUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ṩ(category)renderĳʵ.</zh>
 * <en>Base class providing common services for category renders.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public abstract class AbstractCategoryRender extends AbstractWithAxisRender implements
  Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 2470745792704798948L ;

  /**
   * <zh>ȱʡհռİٷֱ.</zh>
   * <en>The default item margin percentage.</en>
   */
  public static final double DEFAULT_ITEM_MARGIN = 0.20 ;

  /**
   * <zh>ȱʡڻ״ͼĵĻֵ,̶ֵֵСֵֵ֮.</zh>
   * <en>The default base value for the bars,and the value must between the minimum and
   * maximum value of axis tick mark.</en>
   */
  public static final double DEFAULT_BASE_VALUE = 0.0 ;

  /**
   * <zh>ȱʡ״ͼСƿ.Сڴֵ.</zh>
   * <en>The default minimum bar length (in Java2D units).</en>
   */
  public static final double DEFAULT_MINIMUM_BAR_WIDTH = 0.0 ;

  /**
   * <zh>ȱʡÿ״ͼ(λΪJava2D׼λ).</zh>
   * <en>The default maximum bar width (in Java2D units).</en>
   */
  public static final double DEFAULT_MAXIMUM_BAR_WIDTH = Integer.MAX_VALUE ;

  /**
   * <zh>հռİٷֱ.</zh>
   * <en>The margin between items (bars) within a category. </en>
   */
  private double itemMargin ;

  /**
   * <zh>ڻ״ͼĵĻֵ,ֵݼСֵֵ֮.</zh>
   * <en>The base value for the bars.</en>
   */
  private double baseValue ;

  /**
   * <zh>״ͼСƿ.Сڴֵ.</zh>
   * <en>The minimum bar length (in Java2D units).</en>
   */
  private double minimumBarWidth ;

  /**
   * <zh>ÿ״ͼ(λΪJava2D׼λ).</zh>
   * <en>The maximum bar width (in Java2D units).</en>
   */
  private double maximumBarWidth ;

  /**
   * <zh>ע(annotation)󼯺.</zh>
   * <en>The annotation collection.</en>
   */
  private List annotations ;

  /**
   * <zh>.</zh>
   * <en>constructor.</en>
   *
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   */
  protected AbstractCategoryRender(IMatrixDataset dataset) {
    super(dataset) ;

    this.itemMargin = DEFAULT_ITEM_MARGIN ;
    this.baseValue = DEFAULT_BASE_VALUE ;
    this.maximumBarWidth = DEFAULT_MAXIMUM_BAR_WIDTH ;
    this.minimumBarWidth = DEFAULT_MINIMUM_BAR_WIDTH ;

    this.annotations = new ArrayList() ;
  }

  /**
   * <zh>ؿհռİٷֱ.</zh>
   * <en>Returns the item margin as a percentage of the available space for all bars.</en>
   *
   * @return <zh>հռİٷֱ</zh>
   * <en>The margin percentage</en>
   */
  public double getItemMargin() {
    return this.itemMargin ;
  }

  /**
   * <zh>ÿհռİٷֱ,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the item margin and sends a {@link com.ijchart.xychart.event.RenderChangeEvent} to all
   * registered listeners.  The value is expressed as a percentage of the
   * available width for plotting all the bars, with the resulting amount to
   * be distributed between all the bars evenly.</en>
   *
   * @param percent double <zh>հռİٷֱֵ.</zh>
   * <en>The margin (where 0.10 is ten percent).</en>
   */
  public void setItemMargin(double percent) {
    if (this.itemMargin != percent) {
      this.itemMargin = percent ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ڻ״ͼĵĻֵ,ֵݼСֵֵ֮.</zh>
   * <en>The base value for the bars.</en>
   *
   * @return <zh>ڻ״ͼĵĻֵ.</zh>
   * <en>The base value for the bars.</en>
   */
  public double getBaseValue() {
    return this.baseValue ;
  }

  /**
   * <zh>ڻ״ͼĵĻֵ,ֵݼСֵֵ֮,
   * м{@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the base value for the bars and sends a
   *  {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param value double <zh>ڻ״ͼĵĻֵ.</zh>
   * <en>The new base value.</en>
   */
  public void setBaseValue(double value) {
    if (this.baseValue != value) {
      this.baseValue = value ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ÿ״ͼСƿ(λΪJava2D׼λ).</zh>
   * <en>Returns the minimum bar width (in Java2D units).</en>
   *
   * @return double <zh>ÿ״ͼС.</zh>
   * <en>The mimimum bar width.</en>
   */
  public double getMinimumBarWidth() {
    return this.minimumBarWidth ;
  }

  /**
   * <zh>ÿ״ͼС(λΪJava2D׼λ),Сڴֵ.м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the mimimum bar width,if the bar width less than this value would not be drawn
   * ,and sends a {@link com.ijchart.xychart.event.RenderChangeEvent}
   * to all registered listeners.</en>
   *
   * @param barWidth double <zh>ÿ״ͼ(λΪJava2D׼λ).</zh>
   * <en>The minimum bar width (in Java2D units).</en>
   */
  public void setMinimumBarWidth(double barWidth) {
    if (this.minimumBarWidth != barWidth) {
      this.minimumBarWidth = barWidth ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ÿ״ͼ(λΪJava2D׼λ).</zh>
   * <en>Returns the maximum bar width (in Java2D units).</en>
   *
   * @return double <zh>ÿ״ͼ.</zh>
   * <en>The maximum bar width.</en>
   */
  public double getMaximumBarWidth() {
    return this.maximumBarWidth ;
  }

  /**
   * <zh>ÿ״ͼ(λΪJava2D׼λ),м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the maximum bar width,and sends a {@link com.ijchart.xychart.event.RenderChangeEvent}
   * to all registered listeners.</en>
   *
   * @param barWidth double <zh>ÿ״ͼ(λΪJava2D׼λ),0.</zh>
   * <en>The maximum bar width (in Java2D units),and must greater than 0.</en>
   */
  public void setMaximumBarWidth(double barWidth) {
    if (barWidth <= 0) {
      throw new IllegalArgumentException("Must be 'barWidth'>0.") ;
    }

    if (this.maximumBarWidth != barWidth) {
      this.maximumBarWidth = barWidth ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>category marker.</zh>
   * <en>Draws a marker to indicate a particular category on the category axis.</en>
   *
   * @param categoryAxis CategoryAxis <zh>category (Ϊnull).</zh>
   * <en>The category axis (null not permitted).</en>
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼķ (Ϊnull).</zh>
   * <en>The chart orientation (null not permitted).</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   */
  public void drawCategoryMarker(CategoryAxis categoryAxis, Rectangle2D dataArea,
                                 ChartOrientation orientation, RenderShapeCollection renderShapes) {
    if (categoryAxis == null) {
      throw new IllegalArgumentException("Null 'categoryAxis' argument.") ;
    }
    if (dataArea == null) {
      throw new IllegalArgumentException("Null 'dataArea' argument.") ;
    }
    if (orientation == null) {
      throw new IllegalArgumentException("Null 'orientation' argument.") ;
    }
    if (renderShapes == null) {
      throw new IllegalArgumentException("Null 'renderShapes' argument.") ;
    }

    IMatrixDataset dataset = getDataset() ;
    int columns = dataset.getColumnCount() ;
    int rows = dataset.getRowCount() ;
    if (rows == 0 || columns == 0) { //ݿʾ
      return ;
    }

    Collection foregroundMarkers = categoryAxis.getCategoryMarkers(Layer.FOREGROUND) ;
    Collection backgroundMarkers = categoryAxis.getCategoryMarkers(Layer.BACKGROUND) ;

    //״ͼ
    if (orientation.equals(ChartOrientation.VERTICAL)) { //ֱ
      double gapSize = 0 ; //С
      if (rows > 1) {
        gapSize = dataArea.getWidth() * categoryAxis.getCategoryMargin() / (rows - 1) ;
      }
      double lowerMargin = categoryAxis.getLowerMargin() ;
      double upperMargin = categoryAxis.getUpperMargin() ;
      double axisMarginRow = lowerMargin + upperMargin
                             + ((rows == 1) ? 0 : categoryAxis.getCategoryMargin()) ; //еĿհ
      double axisMarginCol = lowerMargin + upperMargin ; //еĿհ
      double categoryWidthRow = dataArea.getWidth() * (1 - axisMarginRow) / rows ; //ÿϵеĿ
      double categoryWidthCol = dataArea.getWidth() * (1 - axisMarginCol) / columns ; //ÿϵеĿ
      double categoryOneMarginRow = (dataArea.getWidth() * (1 - lowerMargin - upperMargin)
                                     * getItemMargin()) / rows ; //ÿcategoryĿհ
      double categoryOneMarginCol = (dataArea.getWidth() * (1 - lowerMargin - upperMargin)
                                     * getItemMargin()) / columns ; //ÿcategoryĿհ
      double barItemMarginRow = categoryOneMarginRow / columns ; //ÿ֮Ŀ
      double barItemMarginCol = categoryOneMarginCol / rows ; //ÿ֮Ŀ

      Iterator iter = foregroundMarkers.iterator() ; //ǰmarker
      while (iter.hasNext()) {
        Object next = iter.next() ;
        if (next instanceof CategoryMarker) {
          CategoryMarker marker = (CategoryMarker) next ;
          if (marker != null) {
            for (int i = 0 ; i < rows ; i++) {
              if (EqualsUtil.equals(marker.getKey(), dataset.getRowKey(i))) {
                float strokeWidth = StrokeUtil.getStrokeWidth(marker.getBorderStroke()) ;
                double startX = dataArea.getX() + dataArea.getWidth() * lowerMargin
                                + i * (categoryWidthRow + gapSize) + 0.5 * barItemMarginRow ;
                Rectangle2D rect = null ;
                if (marker.isDrawAsLine()) {
                  rect = new Rectangle2D.Double(startX
                                                + (categoryWidthRow - barItemMarginRow) * 0.5,
                                                dataArea.getMinY(), strokeWidth, dataArea.getHeight()) ;
                } else {
                  rect = new Rectangle2D.Double(startX, dataArea.getMinY(),
                                                categoryWidthRow - barItemMarginRow,
                                                dataArea.getHeight()) ;
                }
                IChartShape shape = createCategoryMarkerShape(marker, rect) ;
                renderShapes.addForegroundMarkerShape(shape) ;
              }
            }

            for (int i = 0 ; i < columns ; i++) {
              if (EqualsUtil.equals(marker.getKey(), dataset.getColKey(i))) {
                float strokeWidth = StrokeUtil.getStrokeWidth(marker.getBorderStroke()) ;
                double startX = dataArea.getX() + dataArea.getWidth() * lowerMargin
                                + i * (categoryWidthCol + gapSize) + 0.5 * barItemMarginCol ;
                Rectangle2D rect = null ;
                if (marker.isDrawAsLine()) {
                  rect = new Rectangle2D.Double(startX
                                                + (categoryWidthCol - barItemMarginCol) * 0.5,
                                                dataArea.getMinY(), strokeWidth, dataArea.getHeight()) ;
                } else {
                  rect = new Rectangle2D.Double(startX, dataArea.getMinY(),
                                                categoryWidthCol - barItemMarginCol,
                                                dataArea.getHeight()) ;
                }
                IChartShape shape = createCategoryMarkerShape(marker, rect) ;
                renderShapes.addForegroundMarkerShape(shape) ;
              }
            }

          }
        }
      }

      iter = backgroundMarkers.iterator() ; //marker
      while (iter.hasNext()) {
        Object next = iter.next() ;
        if (next instanceof CategoryMarker) {
          CategoryMarker marker = (CategoryMarker) next ;
          if (marker != null) {
            for (int i = 0 ; i < rows ; i++) {
              if (EqualsUtil.equals(marker.getKey(), dataset.getRowKey(i))) {
                float strokeWidth = StrokeUtil.getStrokeWidth(marker.getBorderStroke()) ;
                double startX = dataArea.getX() + dataArea.getWidth() * lowerMargin
                                + i * (categoryWidthRow + gapSize) + 0.5 * barItemMarginRow ;
                Rectangle2D rect = null ;
                if (marker.isDrawAsLine()) {
                  rect = new Rectangle2D.Double(startX
                                                + (categoryWidthRow - barItemMarginRow) * 0.5,
                                                dataArea.getMinY(), strokeWidth, dataArea.getHeight()) ;
                } else {
                  rect = new Rectangle2D.Double(startX, dataArea.getMinY(),
                                                categoryWidthRow - barItemMarginRow,
                                                dataArea.getHeight()) ;
                }
                IChartShape shape = createCategoryMarkerShape(marker, rect) ;
                renderShapes.addBackgroundMarkerShape(shape) ;
              }
            }

            for (int i = 0 ; i < columns ; i++) {
              if (EqualsUtil.equals(marker.getKey(), dataset.getColKey(i))) {
                float strokeWidth = StrokeUtil.getStrokeWidth(marker.getBorderStroke()) ;
                double startX = dataArea.getX() + dataArea.getWidth() * lowerMargin
                                + i * (categoryWidthCol + gapSize) + 0.5 * barItemMarginCol ;
                Rectangle2D rect = null ;
                if (marker.isDrawAsLine()) {
                  rect = new Rectangle2D.Double(startX
                                                + (categoryWidthCol - barItemMarginCol) * 0.5,
                                                dataArea.getMinY(), strokeWidth, dataArea.getHeight()) ;
                } else {
                  rect = new Rectangle2D.Double(startX, dataArea.getMinY(),
                                                categoryWidthCol - barItemMarginCol,
                                                dataArea.getHeight()) ;
                }
                IChartShape shape = createCategoryMarkerShape(marker, rect) ;
                renderShapes.addBackgroundMarkerShape(shape) ;
              }
            }

          }
        }
      }

    } else if (orientation.equals(ChartOrientation.HORIZONTAL)) { //ˮƽ
      double gapSize = 0 ; //С
      if (rows > 1) {
        gapSize = dataArea.getHeight() * categoryAxis.getCategoryMargin() / (rows - 1) ;
      }
      double lowerMargin = categoryAxis.getLowerMargin() ;
      double upperMargin = categoryAxis.getUpperMargin() ;
      double axisMarginRow = lowerMargin + upperMargin
                             + ((rows == 1) ? 0 : categoryAxis.getCategoryMargin()) ; //еĿհ
      double axisMarginCol = lowerMargin + upperMargin ; //еĿհ
      double categoryWidthRow = dataArea.getHeight() * (1 - axisMarginRow) / rows ; //ÿϵеĿ
      double categoryWidthCol = dataArea.getHeight() * (1 - axisMarginCol) / columns ; //ÿϵеĿ
      double categoryOneMarginRow = (dataArea.getHeight() * (1 - lowerMargin - upperMargin)
                                     * getItemMargin()) / rows ; //ÿcategoryĿհ
      double categoryOneMarginCol = (dataArea.getHeight() * (1 - lowerMargin - upperMargin)
                                     * getItemMargin()) / columns ; //ÿcategoryĿհ
      double barItemMarginRow = categoryOneMarginRow / columns ; //ÿ֮Ŀ
      double barItemMarginCol = categoryOneMarginCol / rows ; //ÿ֮Ŀ

      Iterator iter = foregroundMarkers.iterator() ; //ǰmarker
      while (iter.hasNext()) {
        Object next = iter.next() ;
        if (next instanceof CategoryMarker) {
          CategoryMarker marker = (CategoryMarker) next ;
          if (marker != null) {
            for (int i = 0 ; i < rows ; i++) {
              if (EqualsUtil.equals(marker.getKey(), dataset.getRowKey(i))) {
                float strokeWidth = StrokeUtil.getStrokeWidth(marker.getBorderStroke()) ;
                double startY = dataArea.getY() + dataArea.getHeight() * lowerMargin
                                + i * (categoryWidthRow + gapSize) ;
                Rectangle2D rect = null ;
                if (marker.isDrawAsLine()) {
                  rect = new Rectangle2D.Double(dataArea.getMinX(), startY
                                                + categoryWidthRow * 0.5,
                                                dataArea.getWidth(), strokeWidth) ;
                } else {
                  rect = new Rectangle2D.Double(dataArea.getMinX(), startY, dataArea.getWidth(),
                                                categoryWidthRow - barItemMarginRow) ;
                }
                IChartShape shape = createCategoryMarkerShape(marker, rect) ;
                renderShapes.addForegroundMarkerShape(shape) ;
              }
            }

            for (int i = 0 ; i < columns ; i++) {
              if (EqualsUtil.equals(marker.getKey(), dataset.getColKey(i))) {
                float strokeWidth = StrokeUtil.getStrokeWidth(marker.getBorderStroke()) ;
                double startY = dataArea.getY() + dataArea.getHeight() * lowerMargin
                                + i * (categoryWidthCol + gapSize) ;
                Rectangle2D rect = null ;
                if (marker.isDrawAsLine()) {
                  rect = new Rectangle2D.Double(dataArea.getMinX(), startY
                                                + categoryWidthCol * 0.5,
                                                dataArea.getWidth(), strokeWidth) ;
                } else {
                  rect = new Rectangle2D.Double(dataArea.getMinX(), startY, dataArea.getWidth(),
                                                categoryWidthCol - barItemMarginCol) ;
                }
                IChartShape shape = createCategoryMarkerShape(marker, rect) ;
                renderShapes.addForegroundMarkerShape(shape) ;
              }
            }
          }
        }
      }

      iter = backgroundMarkers.iterator() ; //marker
      while (iter.hasNext()) {
        Object next = iter.next() ;
        if (next instanceof CategoryMarker) {
          CategoryMarker marker = (CategoryMarker) next ;
          if (marker != null) {
            for (int i = 0 ; i < rows ; i++) {
              if (EqualsUtil.equals(marker.getKey(), dataset.getRowKey(i))) {
                float strokeWidth = StrokeUtil.getStrokeWidth(marker.getBorderStroke()) ;
                double startY = dataArea.getY()
                                + dataArea.getHeight() * lowerMargin
                                + i * (categoryWidthRow + gapSize) ;
                Rectangle2D rect = null ;
                if (marker.isDrawAsLine()) {
                  rect = new Rectangle2D.Double(dataArea.getMinX(), startY
                                                + categoryWidthRow * 0.5,
                                                dataArea.getWidth(), strokeWidth) ;
                } else {
                  rect = new Rectangle2D.Double(dataArea.getMinX(), startY, dataArea.getWidth(),
                                                categoryWidthRow - barItemMarginRow) ;
                }
                IChartShape shape = createCategoryMarkerShape(marker, rect) ;
                renderShapes.addBackgroundMarkerShape(shape) ;
              }
            }

            for (int i = 0 ; i < columns ; i++) {
              if (EqualsUtil.equals(marker.getKey(), dataset.getRowKey(i))) {
                float strokeWidth = StrokeUtil.getStrokeWidth(marker.getBorderStroke()) ;
                double startY = dataArea.getY()
                                + dataArea.getHeight() * lowerMargin
                                + i * (categoryWidthRow + gapSize) ;
                Rectangle2D rect = null ;
                if (marker.isDrawAsLine()) {
                  rect = new Rectangle2D.Double(dataArea.getMinX(), startY
                                                + categoryWidthRow * 0.5,
                                                dataArea.getWidth(), strokeWidth) ;
                } else {
                  rect = new Rectangle2D.Double(dataArea.getMinX(), startY, dataArea.getWidth(),
                                                categoryWidthRow - barItemMarginRow) ;
                }
                IChartShape shape = createCategoryMarkerShape(marker, rect) ;
                renderShapes.addBackgroundMarkerShape(shape) ;
              }
            }
          }
        }
      }

    }

  }

  /**
   * <zh>category markerͼ.</zh>
   * <en>Create a category marker shape.</en>
   *
   * @param marker CategoryMarker <zh>category marker .</zh>
   * <en>The category marker object.</en>
   * @param rect Rectangle2D <zh>markerͼε.</zh>
   * <en>The marker shape bound.</en>
   *
   * @return IChartShape
   */
  protected static IChartShape createCategoryMarkerShape(CategoryMarker marker, Rectangle2D rect) {
    RectangleTextChartShape shape = new RectangleTextChartShape(rect, marker.getText()) ;

    shape.setFillShape(marker.isFillShape()) ;
    shape.setFillPaint(marker.getFillPaint()) ;
    shape.setDrawBorder(marker.isDrawBorder()) ;
    shape.setBorderPaint(marker.getBorderPaint()) ;
    shape.setBorderStroke(marker.getBorderStroke()) ;
    shape.setFillAlpha(marker.getFillAlpha()) ;
    shape.setTextFont(marker.getTextFont()) ;
    shape.setTextPaint(marker.getTextPaint()) ;
    shape.setDrawText(marker.isDrawText()) ;
    shape.setMargin(marker.getMargin()) ;
    shape.setTextHorizontalAlign(marker.getTextHorizontalAlign()) ;
    shape.setTextVerticalAlign(marker.getTextVerticalAlign()) ;
    shape.setTextPosition(marker.getTextPosition()) ;
    if (marker.isDrawAsLine()) {
      shape.setForceInsideTextNoWrap(true) ;
    }

    return shape ;
  }

  /**
   * <zh>еͼ.</zh>
   * <en>Draws all data items.</en>
   *
   * @param categoryAxis CategoryAxis <zh>category (Ϊnull).</zh>
   * <en>The category axis (null not permitted).</en>
   * @param valueAxis AbstractValueAxis <zh>value (Ϊnull).</zh>
   * <en>The value axis (null not permitted).</en>
   * @param valueAxisEdge RectangleEdge <zh>valueλ (Ϊnull).</zh>
   * <en>The value axis edge (null not permitted).</en>
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼķ (Ϊnull).</zh>
   * <en>The chart orientation (null not permitted).</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   * @param alpha float <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public abstract void drawAllItems(CategoryAxis categoryAxis, AbstractValueAxis valueAxis,
                                    RectangleEdge valueAxisEdge, Rectangle2D dataArea,
                                    ChartOrientation orientation,
                                    RenderShapeCollection renderShapes, float alpha) ;

  /**
   * <zh>һע(annotation),м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Add an annotation and sends a {@link com.ijchart.xychart.event.RenderChangeEvent}
   * to all registered listeners.</en>
   *
   * @param annotation ICategoryAnnotation <zh>ע(annotation) (Ϊnull).</zh>
   * <en>A annotation object (null not permitted).</en>
   */
  public void addAnnotation(ICategoryAnnotation annotation) {
    if (annotation == null) {
      throw new IllegalArgumentException("Null 'annotation' argument.") ;
    }

    this.annotations.add(annotation) ;
    fireRenderChanged() ;
  }

  /**
   * <zh>ɾеע(annotation),м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Remove all annotations of this render and sends a
   *  {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   */
  public void removeAllAnnotation() {
    this.annotations.clear() ;
  }

  /**
   * <zh>еע(annotation),˼ǲܱĵ.</zh>
   * <en>Returns all annotations and this collection is unmodifiable.</en>
   *
   * @return Collection <zh>ע(annotation)󼯺 (Ϊnull).</zh>
   * <en>The collection of annotations (never null).</en>
   */
  public Collection getAllAnnotations() {
    return Collections.unmodifiableCollection(this.annotations) ;
  }

  /**
   * <zh>еע(annotation)ͼ.</zh>
   * <en>Draws all annotation items.</en>
   *
   * @param categoryAxis CategoryAxis <zh>category (Ϊnull).</zh>
   * <en>The category axis (null not permitted).</en>
   * @param valueAxis AbstractValueAxis <zh>value (Ϊnull).</zh>
   * <en>The value axis (null not permitted).</en>
   * @param valueAxisEdge RectangleEdge <zh>valueλ (Ϊnull).</zh>
   * <en>The value axis edge (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼķ (Ϊnull).</zh>
   * <en>The chart orientation (null not permitted).</en>
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param annotationShapes AnnotationShapeCollection <zh>ע(annotation)ͼμ (Ϊnull).</zh>
   * <en>Collects information about the annotation (null not permitted).</en>
   */
  public void drawAllAnnotation(CategoryAxis categoryAxis, AbstractValueAxis valueAxis,
                                RectangleEdge valueAxisEdge, Rectangle2D dataArea,
                                ChartOrientation orientation,
                                AnnotationShapeCollection annotationShapes) {
    for (int i = 0 ; i < this.annotations.size() ; i++) {
      ICategoryAnnotation annotation = (ICategoryAnnotation)this.annotations.get(i) ;
      if (annotation != null) {
        annotation.drawAnnotation(categoryAxis, valueAxis, valueAxisEdge,
                                  dataArea, orientation, annotationShapes) ;
      }
    }
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Indicates if some other object is "equal to" this one.</en>
   *
   * @param obj <zh>ҪȽϵĶ</zh>
   * <en>the reference object with which to compare.</en>
   * @return <zh>ȷtrue,򷵻false.</zh>
   * <en>true if this object is the same as the obj argument; false otherwise.</en>
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true ;
    }
    if (!(obj instanceof AbstractCategoryRender)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    AbstractCategoryRender that = (AbstractCategoryRender) obj ;
    if (this.itemMargin != that.itemMargin) {
      return false ;
    }
    if (this.baseValue != that.baseValue) {
      return false ;
    }
    if (this.minimumBarWidth != that.minimumBarWidth) {
      return false ;
    }
    if (this.maximumBarWidth != that.maximumBarWidth) {
      return false ;
    }
    if (EqualsUtil.equals(this.annotations, that.annotations)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    int result = super.hashCode() ;
    result = 37 * result + (int) (this.itemMargin * 100) ;
    result = 37 * result + (int) (this.baseValue * 100) ;
    result = 37 * result + (int) (this.minimumBarWidth * 100) ;
    result = 37 * result + (int) (this.maximumBarWidth * 100) ;
    result = 37 * result + HashcodeUtil.hashCode(this.annotations) ;

    return result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[itemMargin=" + this.itemMargin
      + ",baseValue=" + this.baseValue
      + ",minimumBarWidth = " + this.minimumBarWidth
      + ",maximumBarWidth=" + this.maximumBarWidth
      + "]" ;
  }

}
