package com.ijchart.xychart.render.category.outlier ;

import java.awt.geom.Point2D ;
import java.io.Serializable ;
import java.util.ArrayList ;
import java.util.Iterator ;
import java.util.List ;

import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>Outlier.</zh>
 * <en>A collection of outliers for a single entity in a box and whisker plot.
 *
 * Outliers are grouped in lists for each entity. Lists contain
 * one or more outliers, determined by whether overlaps have
 * occured. Overlapping outliers are grouped in the same list.
 *
 * Each list contains an averaged outlier, which is the same as a single
 * outlier if there is only one outlier in the list, but the average of
 * all the outliers in the list if there is more than one.
 *
 * NB This is simply my scheme for displaying outliers, and might not be
 * acceptable by the wider community.</en>
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class OutlierList implements Serializable, Cloneable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -4033203004232805293L ;

  /**
   * <zh>Outlier洢.</zh>
   * <en>Storage for the outliers.</en>
   */
  private List outliers ;

  /**
   * <zh>ֵ.</zh>
   * <en>The averaged outlier.</en>
   */
  private Outlier averagedOutlier ;

  /**
   * <zh>Ƿжoutlier.</zh>
   * <en>A flag that indicates whether or not there are multiple outliers in the list.</en>
   */
  private boolean multiple ;

  /**
   * <zh>.</zh>
   * <en>Creates a new list containing a single outlier.</en>
   *
   * @param outlier Outlier <zh>Outlier.</zh>
   * <en>The outlier.</en>
   */
  public OutlierList(Outlier outlier) {
    this.outliers = new ArrayList() ;
    this.multiple = false ;
    setAveragedOutlier(outlier) ;
  }

  /**
   * <zh>һOutlier.</zh>
   * <en>Adds an outlier to the list.</en>
   *
   * @param outlier Outlier <zh>Outlier.</zh>
   * <en>The outlier.</en>
   *
   * @return A boolean.
   */
  public boolean add(Outlier outlier) {
    return this.outliers.add(outlier) ;
  }

  /**
   * <zh>ؼϵ.</zh>
   * <en>Returns the number of outliers in the list.</en>
   *
   * @return int <zh>ϵ.</zh>
   * <en>The item count.</en>
   */
  public int getItemCount() {
    return this.outliers.size() ;
  }

  /**
   * <zh>ƽֵOutlier.</zh>
   * <en>Returns the averaged outlier.</en>
   *
   * @return <zh>ƽֵOutlier.</zh>
   * <en>The averaged outlier.</en>
   */
  public Outlier getAveragedOutlier() {
    return this.averagedOutlier ;
  }

  /**
   * <zh>ƽֵOutlier.</zh>
   * <en>Sets the averaged outlier.</en>
   *
   * @param averagedOutlier Outlier <zh>Outlier.</zh>
   * <en>The averaged outlier.</en>
   */
  public void setAveragedOutlier(Outlier averagedOutlier) {
    this.averagedOutlier = averagedOutlier ;
  }

  /**
   * <zh>ؼǷжoutlier.</zh>
   * <en>Returns true if the list contains multiple outliers, and false otherwise.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isMultiple() {
    return this.multiple ;
  }

  /**
   * <zh>üǷжoutlier.</zh>
   * <en>Sets the flag that indicates whether or not this list represents multiple outliers.</en>
   *
   * @param multiple boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setMultiple(boolean multiple) {
    this.multiple = multiple ;
  }

  /**
   * <zh>Ƿص.</zh>
   * <en>Returns true if the outlier overlaps, and false otherwise.</en>
   *
   * @param other Outlier <zh>Outlier.</zh>
   * <en>The outlier.</en>
   *
   * @return A boolean.
   */
  public boolean isOverlapped(Outlier other) {
    if (other == null) {
      return false ;
    }

    return other.overlaps(getAveragedOutlier()) ;
  }

  /**
   * <zh>ƽֵoutlier.</zh>
   * <en>Updates the averaged outlier.</en>
   */
  public void updateAveragedOutlier() {
    double totalXCoords = 0.0 ;
    double totalYCoords = 0.0 ;
    int size = getItemCount() ;
    for (Iterator iterator = this.outliers.iterator() ;
                             iterator.hasNext() ; ) {
      Outlier o = (Outlier) iterator.next() ;
      totalXCoords += o.getX() ;
      totalYCoords += o.getY() ;
    }

    Point2D point = new Point2D.Double(totalXCoords / size, totalYCoords / size) ;
    getAveragedOutlier().getPoint().setLocation(point) ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof OutlierList)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    OutlierList that = (OutlierList) obj ;
    if (!EqualsUtil.equals(this.outliers, that.outliers)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.averagedOutlier, that.averagedOutlier)) {
      return false ;
    }

    if (this.multiple != that.multiple) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double result = super.hashCode() + HashcodeUtil.hashCode(this.outliers) ;
    result = 37 * result + HashcodeUtil.hashCode(this.averagedOutlier) ;
    result = 37 * result + HashcodeUtil.hashCode(this.multiple) ;

    return (int) result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a textual representation of the outlier.</en>
   *
   * @return A String representing the outlier.
   */
  public String toString() {
    return "[outliers size="
      + ((this.outliers == null) ? "null" : String.valueOf(this.outliers.size()))
      + ",averagedOutlier=" + this.averagedOutlier.toString()
      + ",multiple=" + this.multiple + "]" ;
  }
}
